/*
dice.c Version 0.2.0. Roll a pair of dice
Copyright (C) 2004-2010 dondalah721@yahoo.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

	Free Software Foundation, Inc.
	59 Temple Place - Suite 330
	Boston, MA  02111-1307, USA.
*/

/* to define the escape key */
#define XK_MISCELLANY 1
#define XK_LATIN1 1

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <sys/times.h>
#include <sys/types.h>
#include <X11/Xlib.h>
#include <X11/keysymdef.h>
#include <assert.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "rnd.h"

/* These fonts are defined in dice.mak */
/* Debian font */
/* #define FONT "8x16" */

/* Slackware font */
/* #define FONT "linux8x16" */

#define NIL (0)
#define SCRW 190
#define SCRH 170
#define LEFT   20
#define RIGHT 110
#define CIRCLE 23040

int whiteColor,blackColor;
int leftlean,rghtlean;
int runflg,exposesw;
int bgsw,altbg;
unsigned long maroon,ivory,navy;
int scrw,scrh;
Display *dpy;
Window w;
GC gc;
Font fontid;
Screen *dfltscr;
Colormap cmap;
XColor scrdef,exctdef;

void putstx(pgm)
char *pgm;
   {
   fprintf(stderr,"Usage: %s [-o] [-b] "
      "[-f filename]\n", pgm);
   fprintf(stderr,"Where -o means \"show "
      "the odds\"\n");
   fprintf(stderr,"      -b means \"alternate "
      "the background color\"\n");
   fprintf(stderr,"      -f filename is "
      "the log file name\n");
   fprintf(stderr,"Click the left  mouse button "
      "to re-roll the dice.\n");
   fprintf(stderr,"Click the right mouse button "
      "to exit.\n");
   exit(1);
   } /* putstx */

void redraw()
   {
   XSetForeground(dpy, gc, blackColor);
   } /* redraw */

/* X Windows code is based on: */
/* http://tronche.lri.fr:8000/gui/x/xlib-tutorial/2nd-program-anatomy.html */

void getkey()
   {
   /* after each race: wait for command or exposure */

   XSelectInput(dpy, w,
      KeyPressMask|ExposureMask|ButtonPressMask);

   exposesw = 0;
   while(1)
      {
      int symbol;
      XEvent e;
      XKeyEvent *k;
      XButtonEvent *b;
      XNextEvent(dpy, &e);
      if (e.type == KeyPress)
         {
         k = (XKeyEvent *) &e;
         symbol = XLookupKeysym(k,0);
         if (symbol == XK_Escape || symbol == XK_q)
            {
            runflg = 0;
            break;
            } /* if quit */
	 else
            {
	    if (symbol == XK_Alt_L) continue;
            runflg = 1;
            break;
            } /* if continue */
         } /* if keypress event */
      else if (e.type == ButtonPress)
         {
         b = (XButtonEvent *) &e;
	 if (b->button == 1)
	    {
	    break;
	    } /* if left button pressed */
	 else
	    {
	    runflg = 0;
	    break;
	    } /* if right button pressed */
         } /* if keypress event */
      else if (e.type == Expose)
         {
         exposesw = 1;
         break;
         } /* if expose event */
      } /* wait for window shutdown */
   } /* getkey */

void ifkey()
   {
   int msk;
   int symbol;
   int XCheckMaskEvent();
   XEvent e;
   XKeyEvent *k;
   XButtonEvent *b;

   msk = KeyPressMask|ExposureMask|ButtonPressMask;

   XSelectInput(dpy, w, msk);

   exposesw = 0;
   while (XCheckMaskEvent(dpy, msk, &e))
      {
      if (e.type == KeyPress)
         {
         k = (XKeyEvent *) &e;
         symbol = XLookupKeysym(k,0);
         if (symbol == XK_Escape
	    || symbol == XK_q)
            {
            runflg = 0;
            } /* if quit */
         } /* if keypress event */
      else if (e.type == ButtonPress)
         {
         b = (XButtonEvent *) &e;
	 if (b->button != 1)
	    {
            runflg = 0;
	    } /* if non-left button pressed */
         } /* if buttonpress event */
      else if (e.type == Expose)
         {
         exposesw = 1;
         } /* if expose event */
      } /* if event received */
   } /* ifkey */

void initx()
   {
   int rslt;
   char title[64];

   dpy = XOpenDisplay(NIL);

   if (dpy == NULL)
      {
      fprintf(stderr,"X Windows failure\n");
      exit(1);
      } /* if X Windows is not active */

   assert(dpy);

   dfltscr = XDefaultScreenOfDisplay(dpy);
   if (dfltscr == NULL)
      {
      fprintf(stderr,"XDefaultScreenOfDisplay failed\n");
      perror("XDefaultScreenOfDisplay failed");
      exit(1);
      } /* if error */

   cmap = XDefaultColormapOfScreen(dfltscr);

   rslt = XAllocNamedColor(dpy,cmap,"maroon",
      &scrdef,&exctdef);

   if (rslt < 0)
      {
      fprintf(stderr,"XAllocNamedColor failed\n");
      perror("XAllocNamedColor failed");
      exit(1);
      } /* if error */
   maroon = scrdef.pixel;

   rslt = XAllocNamedColor(dpy,cmap,"navy",
      &scrdef,&exctdef);

   if (rslt < 0)
      {
      fprintf(stderr,"XAllocNamedColor failed\n");
      perror("XAllocNamedColor failed");
      exit(1);
      } /* if error */
   navy = scrdef.pixel;

   rslt = XAllocNamedColor(dpy,cmap,"ivory",
      &scrdef,&exctdef);

   if (rslt < 0)
      {
      fprintf(stderr,"XAllocNamedColor failed\n");
      perror("XAllocNamedColor failed");
      exit(1);
      } /* if error */
   ivory = scrdef.pixel;

   whiteColor = WhitePixel(dpy, DefaultScreen(dpy));
   blackColor = BlackPixel(dpy, DefaultScreen(dpy));

   w = XCreateSimpleWindow(dpy,
      DefaultRootWindow(dpy),
      0, 0, 
      scrw, scrh,
      0, maroon,
      maroon);

   XSelectInput(dpy, w, StructureNotifyMask);

   XMapWindow(dpy, w);

   gc = XCreateGC(dpy, w, 0, NIL);

   fontid = (Font) XLoadFont(dpy,FONT);

   XSetFont(dpy,gc,fontid);

   XSetForeground(dpy, gc, blackColor);

   XSetWindowBorderWidth(dpy, w, 40);

   sprintf(title,"dice");
   XStoreName(dpy,w,title);
   XSetIconName(dpy,w,title);

#ifdef UPPERLEFT

   XMoveWindow(dpy,w,0,0);

#endif

   while(1)
      {
      XEvent e;
      XNextEvent(dpy, &e);
      if (e.type == MapNotify) break;
      } /* wait for window initialization */

   } /* initx */

int opn(fname)
char *fname;
   {
   int hndl;
   hndl = open(fname,
      O_WRONLY|O_CREAT|O_TRUNC,
      S_IRUSR|S_IWUSR);
   if (hndl > 0) return(hndl);
   perror("dice: open error");
   exit(1);
   } /* opn */

void cls(hndl)
int hndl;
   {
   int stat;
   stat = close(hndl);
   if (stat < 0)
      {
      perror("dice: close error");
      exit(1);
      } /* if close error */
   } /* cls */

void wrtlog(left,right,hndl)
int left;
int right;
int hndl;
   {
   int tot;
   int len;
   int wrtlen;
   time_t t;
   struct tm *lcltm;
   char *dtstr;
   char *p;
   char logstr[128];
   time(&t);
   lcltm = (struct tm *) localtime(&t);
   dtstr = (char *) asctime(lcltm);
   p = (char *) dtstr + strlen(dtstr) - 1;
   *p = ' ';   /* remove eol char */
   tot = left + right;
   sprintf(logstr,"%s %d + %d = %2d\n",
      dtstr, left, right, tot);
   len = strlen(logstr);
   wrtlen = write(hndl,logstr,len);
   if (wrtlen == len) return;
   perror("dice: log error");
   exit(1);
   } /* wrtlog */

void shwodds(left,right)
int left;
int right;
   {
   int tot;
   double odds;
   double percent;
   char str[128];
   tot = left + right;
   if (tot < 2 || tot > 12)
      {
      fprintf(stderr,"Dice: Logic error 3\n");
      XCloseDisplay(dpy);
      exit(1);
      } /* logic error 3 */
   switch(tot) {
      case 2:
	 odds = 1.0 / 36.0;
	 percent = (odds * 100.0);
	 sprintf(str,"Odds:    1 in %.1f", 1.0 / odds);
	 XDrawString(dpy,w,gc,20,120,str,strlen(str));
	 sprintf(str,"Percent: %4.1f", percent);
	 XDrawString(dpy,w,gc,20,140,str,strlen(str));
	 break;
      case 3:
	 odds = 2.0 / 36.0;
	 percent = (odds * 100.0);
	 sprintf(str,"Odds:    1 in %.1f", 1.0 / odds);
	 XDrawString(dpy,w,gc,20,120,str,strlen(str));
	 sprintf(str,"Percent: %4.1f", percent);
	 XDrawString(dpy,w,gc,20,140,str,strlen(str));
	 break;
      case 4:
	 odds = 3.0 / 36.0;
	 percent = (odds * 100.0);
	 sprintf(str,"Odds:    1 in %.1f", 1.0 / odds);
	 XDrawString(dpy,w,gc,20,120,str,strlen(str));
	 sprintf(str,"Percent: %4.1f", percent);
	 XDrawString(dpy,w,gc,20,140,str,strlen(str));
	 break;
      case 5:
	 odds = 4.0 / 36.0;
	 percent = (odds * 100.0);
	 sprintf(str,"Odds:    1 in %.1f", 1.0 / odds);
	 XDrawString(dpy,w,gc,20,120,str,strlen(str));
	 sprintf(str,"Percent: %4.1f", percent);
	 XDrawString(dpy,w,gc,20,140,str,strlen(str));
	 break;
      case 6:
	 odds = 5.0 / 36.0;
	 percent = (odds * 100.0);
	 sprintf(str,"Odds:    1 in %.1f", 1.0 / odds);
	 XDrawString(dpy,w,gc,20,120,str,strlen(str));
	 sprintf(str,"Percent: %4.1f", percent);
	 XDrawString(dpy,w,gc,20,140,str,strlen(str));
	 break;
      case 7:
	 odds = 6.0 / 36.0;
	 percent = (odds * 100.0);
	 sprintf(str,"Odds:    1 in %.1f", 1.0 / odds);
	 XDrawString(dpy,w,gc,20,120,str,strlen(str));
	 sprintf(str,"Percent: %4.1f", percent);
	 XDrawString(dpy,w,gc,20,140,str,strlen(str));
	 break;
      case 8:
	 odds = 5.0 / 36.0;
	 percent = (odds * 100.0);
	 sprintf(str,"Odds:    1 in %.1f", 1.0 / odds);
	 XDrawString(dpy,w,gc,20,120,str,strlen(str));
	 sprintf(str,"Percent: %4.1f", percent);
	 XDrawString(dpy,w,gc,20,140,str,strlen(str));
	 break;
      case 9:
	 odds = 4.0 / 36.0;
	 percent = (odds * 100.0);
	 sprintf(str,"Odds:    1 in %.1f", 1.0 / odds);
	 XDrawString(dpy,w,gc,20,120,str,strlen(str));
	 sprintf(str,"Percent: %4.1f", percent);
	 XDrawString(dpy,w,gc,20,140,str,strlen(str));
	 break;
      case 10:
	 odds = 3.0 / 36.0;
	 percent = (odds * 100.0);
	 sprintf(str,"Odds:    1 in %.1f", 1.0 / odds);
	 XDrawString(dpy,w,gc,20,120,str,strlen(str));
	 sprintf(str,"Percent: %4.1f", percent);
	 XDrawString(dpy,w,gc,20,140,str,strlen(str));
	 break;
      case 11:
	 odds = 2.0 / 36.0;
	 percent = (odds * 100.0);
	 sprintf(str,"Odds:    1 in %.1f", 1.0 / odds);
	 XDrawString(dpy,w,gc,20,120,str,strlen(str));
	 sprintf(str,"Percent: %4.1f", percent);
	 XDrawString(dpy,w,gc,20,140,str,strlen(str));
	 break;
      case 12:
	 odds = 1.0 / 36.0;
	 percent = (odds * 100.0);
	 sprintf(str,"Odds:    1 in %.1f", 1.0 / odds);
	 XDrawString(dpy,w,gc,20,120,str,strlen(str));
	 sprintf(str,"Percent: %4.1f", percent);
	 XDrawString(dpy,w,gc,20,140,str,strlen(str));
	 break;
      default:
         fprintf(stderr,"Dice: Logic error 4\n");
         XCloseDisplay(dpy);
	 exit(1);
      } /* switch tot */
   } /* shwodds */

void die1(left)
int left;
   {
   XFillArc(dpy,w,gc,left+22,42,15,15,0,CIRCLE);
   } /* left1 */

void die2(left)
int left;
   {
   int lean;
   if (!exposesw)
      {
      if (left == LEFT)
	 leftlean = rndpwr(1,sd);
      else
         rghtlean = rndpwr(1,sd);
      } /* if not exposed */
   if (left == LEFT)
      lean = leftlean;
   else
      lean = rghtlean;
   if (lean)
      {
      XFillArc(dpy,w,gc,left+05,25,15,15,0,CIRCLE);
      XFillArc(dpy,w,gc,left+40,60,15,15,0,CIRCLE);
      } /* if lean to right */
   else
      {
      XFillArc(dpy,w,gc,left+40,25,15,15,0,CIRCLE);
      XFillArc(dpy,w,gc,left+05,60,15,15,0,CIRCLE);
      } /* if lean to left */
   } /* left2 */

void die3(left)
int left;
   {
   int lean;
   if (!exposesw)
      {
      if (left == LEFT)
	 leftlean = rndpwr(1,sd);
      else
         rghtlean = rndpwr(1,sd);
      } /* if not exposed */
   if (left == LEFT)
      lean = leftlean;
   else
      lean = rghtlean;
   if (lean)
      {
      XFillArc(dpy,w,gc,left+05,25,15,15,0,CIRCLE);
      XFillArc(dpy,w,gc,left+22,42,15,15,0,CIRCLE);
      XFillArc(dpy,w,gc,left+40,60,15,15,0,CIRCLE);
      } /* if lean to right */
   else
      {
      XFillArc(dpy,w,gc,left+05,60,15,15,0,CIRCLE);
      XFillArc(dpy,w,gc,left+22,42,15,15,0,CIRCLE);
      XFillArc(dpy,w,gc,left+40,25,15,15,0,CIRCLE);
      } /* if lean to left */
   } /* left3 */

void die4(left)
int left;
   {
   XFillArc(dpy,w,gc,left+05,25,15,15,0,CIRCLE);
   XFillArc(dpy,w,gc,left+40,25,15,15,0,CIRCLE);
   XFillArc(dpy,w,gc,left+05,60,15,15,0,CIRCLE);
   XFillArc(dpy,w,gc,left+40,60,15,15,0,CIRCLE);
   } /* left4 */

void die5(left)
int left;
   {
   XFillArc(dpy,w,gc,left+05,25,15,15,0,CIRCLE);
   XFillArc(dpy,w,gc,left+40,25,15,15,0,CIRCLE);
   XFillArc(dpy,w,gc,left+05,60,15,15,0,CIRCLE);
   XFillArc(dpy,w,gc,left+40,60,15,15,0,CIRCLE);
   XFillArc(dpy,w,gc,left+22,42,15,15,0,CIRCLE);
   } /* left5 */

void die6(left)
int left;
   {
   int lean;
   if (!exposesw)
      {
      if (left == LEFT)
	 leftlean = rndpwr(1,sd);
      else
         rghtlean = rndpwr(1,sd);
      } /* if not exposed */
   if (left == LEFT)
      lean = leftlean;
   else
      lean = rghtlean;
   if (lean)
      {
      XFillArc(dpy,w,gc,left+05,25,15,15,0,CIRCLE);
      XFillArc(dpy,w,gc,left+05,42,15,15,0,CIRCLE);
      XFillArc(dpy,w,gc,left+05,60,15,15,0,CIRCLE);
      XFillArc(dpy,w,gc,left+40,25,15,15,0,CIRCLE);
      XFillArc(dpy,w,gc,left+40,42,15,15,0,CIRCLE);
      XFillArc(dpy,w,gc,left+40,60,15,15,0,CIRCLE);
      } /* if horizontal */
   else
      {
      XFillArc(dpy,w,gc,left+05,25,15,15,0,CIRCLE);
      XFillArc(dpy,w,gc,left+22,25,15,15,0,CIRCLE);
      XFillArc(dpy,w,gc,left+40,25,15,15,0,CIRCLE);
      XFillArc(dpy,w,gc,left+05,60,15,15,0,CIRCLE);
      XFillArc(dpy,w,gc,left+22,60,15,15,0,CIRCLE);
      XFillArc(dpy,w,gc,left+40,60,15,15,0,CIRCLE);
      } /* if horizontal */
   } /* left6 */

int main(argc,argv)
int argc;
char **argv;
   {
   int i,j;
   int rslt1,rslt2,rslt3;
   int oddsw,logsw;
   int hndl;
   int bgcolor;
   char fname[256];
   i = j = exposesw = oddsw = bgsw = logsw = 0;
   hndl = 0;
   altbg = 0;
   scrw = SCRW;
   scrh = SCRH - 60;
   if (argc > 5) putstx(*argv);
   if (argc > 1)
      {
      i = 1;
      while (i < argc)
	 {
	 rslt1 = strcmp(*(argv+i),"-o");
	 rslt2 = strcmp(*(argv+i),"-b");
	 rslt3 = strcmp(*(argv+i),"-f");
	 if (!rslt1)
	    {
	    oddsw = 1;
	    scrh = SCRH;
	    } /* if odds calculation */
	 else if (!rslt2)
	    {
	    bgsw = 1;
	    } /* if alternate background */
	 else if (!rslt3)
	    {
	    logsw = 1;
	    i = i + 1;
	    strcpy(fname,*(argv+i));
	    hndl = opn(fname);
	    } /* if logging file */
	 else
	    {
	    fprintf(stderr,"dice: invalid parm %s\n",
	       *(argv+i));
	    putstx(*argv);
	    exit(1);
	    } /* bad parm */
	 i++;
	 } /* for each parm */
      } /* if parm present */
   sd = (unsigned char *) rndinit();
   initx();
   bgcolor = maroon;
   runflg = 1;
   while (runflg)
      {
      if (exposesw)
	 {
         XSetForeground(dpy, gc, bgcolor);
         XSetBackground(dpy, gc, bgcolor);
         XFillRectangle(dpy,w,gc,0,0,scrw,scrh);
         XSetForeground(dpy, gc, ivory);
	 } /* if exposed */
      if (bgsw && !exposesw)
         {
	 if (altbg)
	    {
	    bgcolor = maroon;
            XSetForeground(dpy, gc, bgcolor);
            XSetBackground(dpy, gc, bgcolor);
            XFillRectangle(dpy,w,gc,0,0,scrw,scrh);
            XSetForeground(dpy, gc, ivory);
	    } /* if maroon background */
	 else
	    {
	    bgcolor = navy;
            XSetForeground(dpy, gc, bgcolor);
            XSetBackground(dpy, gc, bgcolor);
            XFillRectangle(dpy,w,gc,0,0,scrw,scrh);
            XSetForeground(dpy, gc, ivory);
	    } /* if navy background */
	 altbg = (altbg ^ 1) & 1;
	 } /* if alternate background */
      XSetForeground(dpy, gc, ivory);
      XFillRectangle(dpy,w,gc,LEFT,20,60,60);
      XFillRectangle(dpy,w,gc,RIGHT,20,60,60);
      XSetForeground(dpy, gc, blackColor);
      if (!exposesw)
	 {
         i = rndnum(6,sd) + 1;
	 } /* if not exposed */
      switch(i) {
         case 1:
	    die1(LEFT);
	    break;
         case 2:
	    die2(LEFT);
	    break;
         case 3:
	    die3(LEFT);
	    break;
         case 4:
	    die4(LEFT);
	    break;
         case 5:
	    die5(LEFT);
	    break;
         case 6:
	    die6(LEFT);
	    break;
         default:
            fprintf(stderr,"Dice: Logic error 1\n");
	    exit(0);
         } /* switch i */
      if (!exposesw)
	 {
         j = rndnum(6,sd) + 1;
	 } /* if not exposed */
      switch(j) {
         case 1:
	    die1(RIGHT);
	    break;
         case 2:
	    die2(RIGHT);
	    break;
         case 3:
	    die3(RIGHT);
	    break;
         case 4:
	    die4(RIGHT);
	    break;
         case 5:
	    die5(RIGHT);
	    break;
         case 6:
	    die6(RIGHT);
	    break;
         default:
            fprintf(stderr,"Dice: Logic error 2\n");
	    exit(0);
         } /* switch j */
      if (oddsw)
	 {
	 /* the switch has already been toggled */
         XSetBackground(dpy, gc, bgcolor);
         XSetForeground(dpy, gc, bgcolor);
         XFillRectangle(dpy, w, gc, 20, 100, 160, 40);
         XSetForeground(dpy, gc, ivory);
	 shwodds(i,j);
	 } /* if show odds */
      if (!exposesw)
	 {
	 if (logsw) wrtlog(i,j,hndl);
	 } /* if not exposed */
      getkey();
      } /* for each roll */
   XCloseDisplay(dpy);
   if (logsw) cls(hndl);
   return(0);
   } /* main */
